"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _stream = _interopRequireDefault(require("stream"));

var _os = require("os");

var _constants = require("./constants");

var _resourceErrors = require("./resource-errors");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const EventStream = function ({
  pool,
  maintenanceMode,
  byteOffset
}) {
  _stream.default.Writable.call(this, {
    objectMode: true
  });

  this.pool = pool;
  this.byteOffset = byteOffset;
  this.buffer = Buffer.allocUnsafe(_constants.BUFFER_SIZE);
  this.beginPosition = 0;
  this.endPosition = 0;
  this.vacantSize = _constants.BUFFER_SIZE;
  this.saveEventPromiseSet = new Set();
  this.saveEventErrors = [];
  this.timestamp = 0;
  this.maintenanceMode = maintenanceMode;
  this.isMaintenanceInProgress = false;
  this.parsedEventsCount = 0;
  this.bypassMode = false;
  this.savedEventsCount = 0;
  this.on('timeout', () => {
    this.externalTimeout = true;
  });
};

EventStream.prototype = Object.create(_stream.default.Writable.prototype);
EventStream.prototype.constructor = _stream.default.Writable;

EventStream.prototype._write = async function (chunk, encoding, callback) {
  if (this.bypassMode) {
    await new Promise(resolve => setImmediate(resolve));
    callback();
    return;
  }

  try {
    await this.pool.waitConnect();
    const {
      dropEvents,
      initEvents,
      freeze,
      injectEvent
    } = this.pool;

    if (this.maintenanceMode === _constants.MAINTENANCE_MODE_AUTO && this.isMaintenanceInProgress === false) {
      this.isMaintenanceInProgress = true;

      try {
        await dropEvents();
      } catch (error) {
        if (!_resourceErrors.ResourceNotExistError.is(error)) {
          throw error;
        }
      }

      await initEvents();
      await freeze();
    }

    if (this.encoding == null) {
      this.encoding = encoding;
    } else if (this.encoding !== encoding) {
      throw new Error('Multiple encodings not supported');
    }

    if (chunk.byteLength > this.vacantSize) {
      throw new Error('Buffer overflow');
    }

    if (chunk.byteLength + this.endPosition <= _constants.BUFFER_SIZE) {
      chunk.copy(this.buffer, this.endPosition);
    } else {
      chunk.copy(this.buffer, this.endPosition, 0, _constants.BUFFER_SIZE - this.endPosition);
      chunk.copy(this.buffer, 0, _constants.BUFFER_SIZE - this.endPosition, chunk.byteLength);
    }

    this.endPosition = (this.endPosition + chunk.byteLength) % _constants.BUFFER_SIZE;
    this.vacantSize -= chunk.byteLength;

    if (this.vacantSize === _constants.BUFFER_SIZE) {
      callback();
      return;
    }

    let eolPosition = 0;

    while (true) {
      eolPosition = chunk.indexOf('\n', eolPosition, this.encoding);

      if (eolPosition < 0) {
        break;
      } else {
        eolPosition++;
      }

      const endEventPosition = (_constants.BUFFER_SIZE + this.endPosition - chunk.byteLength + eolPosition) % _constants.BUFFER_SIZE;
      let stringifiedEvent = null;
      let eventByteLength = 0;

      if (this.beginPosition < endEventPosition) {
        stringifiedEvent = this.buffer.slice(this.beginPosition, endEventPosition).toString(this.encoding);
        eventByteLength += endEventPosition - this.beginPosition;
      } else {
        stringifiedEvent = this.buffer.slice(this.beginPosition, _constants.BUFFER_SIZE).toString(this.encoding);
        stringifiedEvent += this.buffer.slice(0, endEventPosition).toString(this.encoding);
        eventByteLength += _constants.BUFFER_SIZE - this.beginPosition + endEventPosition;
      }

      this.vacantSize += eventByteLength;
      this.beginPosition = endEventPosition;
      this.byteOffset += eventByteLength;
      const event = JSON.parse(stringifiedEvent);
      this.timestamp = Math.max(this.timestamp, event.timestamp);
      const saveEventPromise = injectEvent(event).then(() => {
        this.savedEventsCount++;
      }).catch(this.saveEventErrors.push.bind(this.saveEventErrors));
      void saveEventPromise.then(this.saveEventPromiseSet.delete.bind(this.saveEventPromiseSet, saveEventPromise));
      this.saveEventPromiseSet.add(saveEventPromise);

      if (this.parsedEventsCount++ >= _constants.BATCH_SIZE) {
        await Promise.all([...this.saveEventPromiseSet]);

        if (this.externalTimeout === true) {
          this.bypassMode = true;
        }

        this.parsedEventsCount = 0;
      }
    }

    callback();
  } catch (error) {
    callback(error);
  }
};

EventStream.prototype._final = async function (callback) {
  if (this.bypassMode) {
    try {
      await Promise.all([...this.saveEventPromiseSet]);
      callback();
    } catch (err) {
      callback(err);
    } finally {
      this.buffer = null;
    }

    return;
  }

  try {
    await this.pool.waitConnect();
    const {
      unfreeze,
      injectEvent
    } = this.pool;

    if (this.vacantSize !== _constants.BUFFER_SIZE) {
      let stringifiedEvent = null;
      let eventByteLength = 0;

      if (this.beginPosition < this.endPosition) {
        stringifiedEvent = this.buffer.slice(this.beginPosition, this.endPosition).toString(this.encoding);
        eventByteLength += this.endPosition - this.beginPosition;
      } else {
        stringifiedEvent = this.buffer.slice(this.beginPosition, _constants.BUFFER_SIZE).toString(this.encoding);
        stringifiedEvent += this.buffer.slice(0, this.endPosition).toString(this.encoding);
        eventByteLength += _constants.BUFFER_SIZE - this.beginPosition + this.endPosition;
      }

      let event = _constants.PARTIAL_EVENT_FLAG;

      try {
        event = JSON.parse(stringifiedEvent);
      } catch (_unused) {}

      if (event !== _constants.PARTIAL_EVENT_FLAG) {
        this.timestamp = Math.max(this.timestamp, event.timestamp);
        this.byteOffset += eventByteLength;
        const saveEventPromise = injectEvent(event).then(() => {
          this.savedEventsCount++;
        }).catch(this.saveEventErrors.push.bind(this.saveEventErrors));
        void saveEventPromise.then(this.saveEventPromiseSet.delete.bind(this.saveEventPromiseSet, saveEventPromise));
        this.saveEventPromiseSet.add(saveEventPromise);
      }
    }

    await Promise.all([...this.saveEventPromiseSet]);

    if (this.maintenanceMode === _constants.MAINTENANCE_MODE_AUTO && this.isMaintenanceInProgress === true) {
      this.isMaintenanceInProgress = false;
      await unfreeze();
    }

    if (this.saveEventErrors.length > 0) {
      const error = new Error(this.saveEventErrors.map(({
        message
      }) => message).join(_os.EOL));
      error.stack = this.saveEventErrors.map(({
        stack
      }) => stack).join(_os.EOL);
      throw error;
    }

    callback();
  } catch (error) {
    callback(error);
  } finally {
    this.buffer = null;
  }
};

const importEventsStream = (pool, {
  byteOffset = 0,
  maintenanceMode = _constants.MAINTENANCE_MODE_AUTO
} = {}) => {
  switch (maintenanceMode) {
    case _constants.MAINTENANCE_MODE_AUTO:
    case _constants.MAINTENANCE_MODE_MANUAL:
      return new EventStream({
        pool,
        maintenanceMode,
        byteOffset
      });

    default:
      throw new Error(`Wrong maintenance mode ${String(maintenanceMode)}`);
  }
};

var _default = importEventsStream;
exports.default = _default;
//# sourceMappingURL=import-events.js.map